@@ -8,18 +8,20 @@ import sys
 import asyncio
 import logging
 from types import TracebackType
-from typing import Any, AsyncGenerator, AsyncIterator, Dict, Optional
-from typing import Type, Union, cast, overload, TYPE_CHECKING
+from typing import Any, AsyncGenerator, AsyncIterator, Dict, List, Optional
+from typing import Type, TypeVar, Union, cast, overload, TYPE_CHECKING
+from contextlib import asynccontextmanager
 
+from . import pq
 from . import errors as e
 from . import waiting
-from .pq import Format
 from .abc import AdaptContext, Params, PQGen, PQGenConn, Query, RV
-from .rows import Row, AsyncRowFactory, tuple_row, TupleRow
+from ._tpc import Xid
+from .rows import Row, AsyncRowFactory, tuple_row, TupleRow, args_row
 from .adapt import AdaptersMap
 from ._enums import IsolationLevel
-from ._compat import asynccontextmanager, get_running_loop
-from .conninfo import make_conninfo, conninfo_to_dict
+from .conninfo import make_conninfo, conninfo_to_dict, resolve_hostaddr_async
+from ._pipeline import AsyncPipeline
 from ._encodings import pgconn_encoding
 from .connection import BaseConnection, CursorRow, Notify
 from .generators import notifies
@@ -30,6 +32,11 @@ from .server_cursor import AsyncServerCursor
 if TYPE_CHECKING:
     from .pq.abc import PGconn
 
+TEXT = pq.Format.TEXT
+BINARY = pq.Format.BINARY
+
+IDLE = pq.TransactionStatus.IDLE
+INTRANS = pq.TransactionStatus.INTRANS
 
 logger = logging.getLogger("psycopg")
 
@@ -44,14 +51,16 @@ class AsyncConnection(BaseConnection[Row]):
     cursor_factory: Type[AsyncCursor[Row]]
     server_cursor_factory: Type[AsyncServerCursor[Row]]
     row_factory: AsyncRowFactory[Row]
+    _pipeline: Optional[AsyncPipeline]
+    _Self = TypeVar("_Self", bound="AsyncConnection[Row]")
 
     def __init__(
         self,
         pgconn: "PGconn",
-        row_factory: Optional[AsyncRowFactory[Row]] = None,
+        row_factory: AsyncRowFactory[Row] = cast(AsyncRowFactory[Row], tuple_row),
     ):
         super().__init__(pgconn)
-        self.row_factory = row_factory or cast(AsyncRowFactory[Row], tuple_row)
+        self.row_factory = row_factory
         self.lock = asyncio.Lock()
         self.cursor_factory = AsyncCursor
         self.server_cursor_factory = AsyncServerCursor
@@ -63,10 +72,13 @@ class AsyncConnection(BaseConnection[Row]):
         conninfo: str = "",
         *,
         autocommit: bool = False,
+        prepare_threshold: Optional[int] = 5,
         row_factory: AsyncRowFactory[Row],
+        cursor_factory: Optional[Type[AsyncCursor[Row]]] = None,
         context: Optional[AdaptContext] = None,
         **kwargs: Union[None, int, str],
     ) -> "AsyncConnection[Row]":
+        # TODO: returned type should be _Self. See #308.
         ...
 
     @overload
@@ -76,6 +88,8 @@ class AsyncConnection(BaseConnection[Row]):
         conninfo: str = "",
         *,
         autocommit: bool = False,
+        prepare_threshold: Optional[int] = 5,
+        cursor_factory: Optional[Type[AsyncCursor[Any]]] = None,
         context: Optional[AdaptContext] = None,
         **kwargs: Union[None, int, str],
     ) -> "AsyncConnection[TupleRow]":
@@ -87,13 +101,15 @@ class AsyncConnection(BaseConnection[Row]):
         conninfo: str = "",
         *,
         autocommit: bool = False,
+        prepare_threshold: Optional[int] = 5,
         context: Optional[AdaptContext] = None,
         row_factory: Optional[AsyncRowFactory[Row]] = None,
+        cursor_factory: Optional[Type[AsyncCursor[Row]]] = None,
         **kwargs: Any,
     ) -> "AsyncConnection[Any]":
 
         if sys.platform == "win32":
-            loop = get_running_loop()
+            loop = asyncio.get_running_loop()
             if isinstance(loop, asyncio.ProactorEventLoop):
                 raise e.InterfaceError(
                     "Psycopg cannot use the 'ProactorEventLoop' to run in async"
@@ -115,11 +131,14 @@ class AsyncConnection(BaseConnection[Row]):
 
         if row_factory:
             rv.row_factory = row_factory
+        if cursor_factory:
+            rv.cursor_factory = cursor_factory
         if context:
             rv._adapters = AdaptersMap(context.adapters)
+        rv.prepare_threshold = prepare_threshold
         return rv
 
-    async def __aenter__(self) -> "AsyncConnection[Row]":
+    async def __aenter__(self: _Self) -> _Self:
         return self
 
     async def __aexit__(
@@ -153,7 +172,15 @@ class AsyncConnection(BaseConnection[Row]):
     async def _get_connection_params(
         cls, conninfo: str, **kwargs: Any
     ) -> Dict[str, Any]:
-        """Manipulate connection parameters before connecting."""
+        """Manipulate connection parameters before connecting.
+
+        .. versionchanged:: 3.1
+            Unlike the sync counterpart, perform non-blocking address
+            resolution and populate the ``hostaddr`` connection parameter,
+            unless the user has provided one themselves. See
+            `~psycopg._dns.resolve_hostaddr_async()` for details.
+
+        """
         params = conninfo_to_dict(conninfo, **kwargs)
 
         # Make sure there is an usable connect_timeout
@@ -162,6 +189,9 @@ class AsyncConnection(BaseConnection[Row]):
         else:
             params["connect_timeout"] = None
 
+        # Resolve host addresses in non-blocking way
+        params = await resolve_hostaddr_async(params)
+
         return params
 
     async def close(self) -> None:
@@ -233,7 +263,7 @@ class AsyncConnection(BaseConnection[Row]):
             cur = self.cursor_factory(self, row_factory=row_factory)
 
         if binary:
-            cur.format = Format.BINARY
+            cur.format = BINARY
 
         return cur
 
@@ -248,7 +278,7 @@ class AsyncConnection(BaseConnection[Row]):
         try:
             cur = self.cursor()
             if binary:
-                cur.format = Format.BINARY
+                cur.format = BINARY
 
             return await cur.execute(query, params, prepare=prepare)
 
@@ -275,27 +305,65 @@ class AsyncConnection(BaseConnection[Row]):
         :rtype: AsyncTransaction
         """
         tx = AsyncTransaction(self, savepoint_name, force_rollback)
-        async with tx:
-            yield tx
+        if self._pipeline:
+            async with self.pipeline(), tx, self.pipeline():
+                yield tx
+        else:
+            async with tx:
+                yield tx
 
     async def notifies(self) -> AsyncGenerator[Notify, None]:
-        while 1:
+        while True:
             async with self.lock:
-                ns = await self.wait(notifies(self.pgconn))
+                try:
+                    ns = await self.wait(notifies(self.pgconn))
+                except e.Error as ex:
+                    raise ex.with_traceback(None)
             enc = pgconn_encoding(self.pgconn)
             for pgn in ns:
-                n = Notify(
-                    pgn.relname.decode(enc), pgn.extra.decode(enc), pgn.be_pid
-                )
+                n = Notify(pgn.relname.decode(enc), pgn.extra.decode(enc), pgn.be_pid)
                 yield n
 
+    @asynccontextmanager
+    async def pipeline(self) -> AsyncIterator[AsyncPipeline]:
+        """Context manager to switch the connection into pipeline mode."""
+        async with self.lock:
+            self._check_connection_ok()
+
+            pipeline = self._pipeline
+            if pipeline is None:
+                # WARNING: reference loop, broken ahead.
+                pipeline = self._pipeline = AsyncPipeline(self)
+
+        try:
+            async with pipeline:
+                yield pipeline
+        finally:
+            if pipeline.level == 0:
+                async with self.lock:
+                    assert pipeline is self._pipeline
+                    self._pipeline = None
+
     async def wait(self, gen: PQGen[RV]) -> RV:
-        return await waiting.wait_async(gen, self.pgconn.socket)
+        try:
+            return await waiting.wait_async(gen, self.pgconn.socket)
+        except KeyboardInterrupt:
+            # TODO: this doesn't seem to work as it does for sync connections
+            # see tests/test_concurrency_async.py::test_ctrl_c
+            # In the test, the code doesn't reach this branch.
+
+            # On Ctrl-C, try to cancel the query in the server, otherwise
+            # otherwise the connection will be stuck in ACTIVE state
+            c = self.pgconn.get_cancel()
+            c.cancel()
+            try:
+                await waiting.wait_async(gen, self.pgconn.socket)
+            except e.QueryCanceled:
+                pass  # as expected
+            raise
 
     @classmethod
-    async def _wait_conn(
-        cls, gen: PQGenConn[RV], timeout: Optional[int]
-    ) -> RV:
+    async def _wait_conn(cls, gen: PQGenConn[RV], timeout: Optional[int]) -> RV:
         return await waiting.wait_conn_async(gen, timeout)
 
     def _set_autocommit(self, value: bool) -> None:
@@ -304,17 +372,15 @@ class AsyncConnection(BaseConnection[Row]):
     async def set_autocommit(self, value: bool) -> None:
         """Async version of the `~Connection.autocommit` setter."""
         async with self.lock:
-            super()._set_autocommit(value)
+            await self.wait(self._set_autocommit_gen(value))
 
     def _set_isolation_level(self, value: Optional[IsolationLevel]) -> None:
         self._no_set_async("isolation_level")
 
-    async def set_isolation_level(
-        self, value: Optional[IsolationLevel]
-    ) -> None:
+    async def set_isolation_level(self, value: Optional[IsolationLevel]) -> None:
         """Async version of the `~Connection.isolation_level` setter."""
         async with self.lock:
-            super()._set_isolation_level(value)
+            await self.wait(self._set_isolation_level_gen(value))
 
     def _set_read_only(self, value: Optional[bool]) -> None:
         self._no_set_async("read_only")
@@ -322,7 +388,7 @@ class AsyncConnection(BaseConnection[Row]):
     async def set_read_only(self, value: Optional[bool]) -> None:
         """Async version of the `~Connection.read_only` setter."""
         async with self.lock:
-            super()._set_read_only(value)
+            await self.wait(self._set_read_only_gen(value))
 
     def _set_deferrable(self, value: Optional[bool]) -> None:
         self._no_set_async("deferrable")
@@ -330,10 +396,41 @@ class AsyncConnection(BaseConnection[Row]):
     async def set_deferrable(self, value: Optional[bool]) -> None:
         """Async version of the `~Connection.deferrable` setter."""
         async with self.lock:
-            super()._set_deferrable(value)
+            await self.wait(self._set_deferrable_gen(value))
 
     def _no_set_async(self, attribute: str) -> None:
         raise AttributeError(
             f"'the {attribute!r} property is read-only on async connections:"
             f" please use 'await .set_{attribute}()' instead."
         )
+
+    async def tpc_begin(self, xid: Union[Xid, str]) -> None:
+        async with self.lock:
+            await self.wait(self._tpc_begin_gen(xid))
+
+    async def tpc_prepare(self) -> None:
+        try:
+            async with self.lock:
+                await self.wait(self._tpc_prepare_gen())
+        except e.ObjectNotInPrerequisiteState as ex:
+            raise e.NotSupportedError(str(ex)) from None
+
+    async def tpc_commit(self, xid: Union[Xid, str, None] = None) -> None:
+        async with self.lock:
+            await self.wait(self._tpc_finish_gen("commit", xid))
+
+    async def tpc_rollback(self, xid: Union[Xid, str, None] = None) -> None:
+        async with self.lock:
+            await self.wait(self._tpc_finish_gen("rollback", xid))
+
+    async def tpc_recover(self) -> List[Xid]:
+        self._check_tpc()
+        status = self.info.transaction_status
+        async with self.cursor(row_factory=args_row(Xid._from_record)) as cur:
+            await cur.execute(Xid._get_recover_query())
+            res = await cur.fetchall()
+
+        if status == IDLE and self.info.transaction_status == INTRANS:
+            await self.rollback()
+
+        return res
