@@ -6,36 +6,49 @@ psycopg async cursor objects
 
 from types import TracebackType
 from typing import Any, AsyncIterator, Iterable, List
-from typing import Optional, Type, TypeVar, TYPE_CHECKING
+from typing import Optional, Type, TypeVar, TYPE_CHECKING, overload
+from contextlib import asynccontextmanager
 
+from . import pq
 from . import errors as e
-
 from .abc import Query, Params
-from .copy import AsyncCopy
+from .copy import AsyncCopy, AsyncWriter as AsyncCopyWriter
 from .rows import Row, RowMaker, AsyncRowFactory
 from .cursor import BaseCursor
-from ._compat import asynccontextmanager
+from ._pipeline import Pipeline
 
 if TYPE_CHECKING:
     from .connection_async import AsyncConnection
 
-_C = TypeVar("_C", bound="AsyncCursor[Any]")
-
 
 class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
     __module__ = "psycopg"
     __slots__ = ()
+    _Self = TypeVar("_Self", bound="AsyncCursor[Row]")
+
+    @overload
+    def __init__(self: "AsyncCursor[Row]", connection: "AsyncConnection[Row]"):
+        ...
+
+    @overload
+    def __init__(
+        self: "AsyncCursor[Row]",
+        connection: "AsyncConnection[Any]",
+        *,
+        row_factory: AsyncRowFactory[Row],
+    ):
+        ...
 
     def __init__(
         self,
         connection: "AsyncConnection[Any]",
         *,
-        row_factory: AsyncRowFactory[Row],
+        row_factory: Optional[AsyncRowFactory[Row]] = None,
     ):
         super().__init__(connection)
-        self._row_factory = row_factory
+        self._row_factory = row_factory or connection.row_factory
 
-    async def __aenter__(self: _C) -> _C:
+    async def __aenter__(self: _Self) -> _Self:
         return self
 
     async def __aexit__(
@@ -63,13 +76,13 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
         return self._row_factory(self)
 
     async def execute(
-        self: _C,
+        self: _Self,
         query: Query,
         params: Optional[Params] = None,
         *,
         prepare: Optional[bool] = None,
         binary: Optional[bool] = None,
-    ) -> _C:
+    ) -> _Self:
         try:
             async with self._conn.lock:
                 await self._conn.wait(
@@ -79,10 +92,33 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
             raise ex.with_traceback(None)
         return self
 
-    async def executemany(self, query: Query, params_seq: Iterable[Params]) -> None:
+    async def executemany(
+        self,
+        query: Query,
+        params_seq: Iterable[Params],
+        *,
+        returning: bool = False,
+    ) -> None:
         try:
-            async with self._conn.lock:
-                await self._conn.wait(self._executemany_gen(query, params_seq))
+            if Pipeline.is_supported():
+                # If there is already a pipeline, ride it, in order to avoid
+                # sending unnecessary Sync.
+                async with self._conn.lock:
+                    p = self._conn._pipeline
+                    if p:
+                        await self._conn.wait(
+                            self._executemany_gen_pipeline(query, params_seq, returning)
+                        )
+                # Otherwise, make a new one
+                if not p:
+                    async with self._conn.pipeline(), self._conn.lock:
+                        await self._conn.wait(
+                            self._executemany_gen_pipeline(query, params_seq, returning)
+                        )
+            else:
+                await self._conn.wait(
+                    self._executemany_gen_no_pipeline(query, params_seq, returning)
+                )
         except e.Error as ex:
             raise ex.with_traceback(None)
 
@@ -93,16 +129,33 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
         *,
         binary: Optional[bool] = None,
     ) -> AsyncIterator[Row]:
-        async with self._conn.lock:
-            await self._conn.wait(self._stream_send_gen(query, params, binary=binary))
-            first = True
-            while await self._conn.wait(self._stream_fetchone_gen(first)):
-                rec = self._tx.load_row(0, self._make_row)
-                assert rec is not None
-                yield rec
-                first = False
+        if self._pgconn.pipeline_status:
+            raise e.ProgrammingError("stream() cannot be used in pipeline mode")
+
+        try:
+            async with self._conn.lock:
+                await self._conn.wait(
+                    self._stream_send_gen(query, params, binary=binary)
+                )
+                first = True
+                while await self._conn.wait(self._stream_fetchone_gen(first)):
+                    # We know that, if we got a result, it has a single row.
+                    rec: Row = self._tx.load_row(0, self._make_row)  # type: ignore
+                    yield rec
+                    first = False
+        except e.Error as ex:
+            # try to get out of ACTIVE state. Just do a single attempt, which
+            # should work to recover from an error or query cancelled.
+            if self._pgconn.transaction_status == pq.TransactionStatus.ACTIVE:
+                try:
+                    await self._conn.wait(self._stream_fetchone_gen(first))
+                except Exception:
+                    pass
+
+            raise ex.with_traceback(None)
 
     async def fetchone(self) -> Optional[Row]:
+        await self._fetch_pipeline()
         self._check_result_for_fetch()
         rv = self._tx.load_row(self._pos, self._make_row)
         if rv is not None:
@@ -110,6 +163,7 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
         return rv
 
     async def fetchmany(self, size: int = 0) -> List[Row]:
+        await self._fetch_pipeline()
         self._check_result_for_fetch()
         assert self.pgresult
 
@@ -124,6 +178,7 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
         return records
 
     async def fetchall(self) -> List[Row]:
+        await self._fetch_pipeline()
         self._check_result_for_fetch()
         assert self.pgresult
         records = self._tx.load_rows(self._pos, self.pgresult.ntuples, self._make_row)
@@ -131,12 +186,13 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
         return records
 
     async def __aiter__(self) -> AsyncIterator[Row]:
+        await self._fetch_pipeline()
         self._check_result_for_fetch()
 
         def load(pos: int) -> Optional[Row]:
             return self._tx.load_row(pos, self._make_row)
 
-        while 1:
+        while True:
             row = load(self._pos)
             if row is None:
                 break
@@ -147,12 +203,32 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
         self._scroll(value, mode)
 
     @asynccontextmanager
-    async def copy(self, statement: Query) -> AsyncIterator[AsyncCopy]:
+    async def copy(
+        self,
+        statement: Query,
+        params: Optional[Params] = None,
+        *,
+        writer: Optional[AsyncCopyWriter] = None,
+    ) -> AsyncIterator[AsyncCopy]:
         """
         :rtype: AsyncCopy
         """
-        async with self._conn.lock:
-            await self._conn.wait(self._start_copy_gen(statement))
+        try:
+            async with self._conn.lock:
+                await self._conn.wait(self._start_copy_gen(statement, params))
 
-        async with AsyncCopy(self) as copy:
-            yield copy
+            async with AsyncCopy(self, writer=writer) as copy:
+                yield copy
+        except e.Error as ex:
+            raise ex.with_traceback(None)
+
+        self._select_current_result(0)
+
+    async def _fetch_pipeline(self) -> None:
+        if (
+            self._execmany_returning is not False
+            and not self.pgresult
+            and self._conn._pipeline
+        ):
+            async with self._conn.lock:
+                await self._conn.wait(self._conn._pipeline._fetch_gen(flush=True))
